//******************************************************************************
// Copyright (c) 2014 Dave Hein
// See end of file for terms of use.
//******************************************************************************
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* INLINE SPIN
CON
  _clkmode = xinput
  _xinfreq = $4e495053 ' SPIN
*/
#ifdef CSPIN_FLAG
#define FILE long
#define SEEK_SET 0
#endif

// SPIN OBJECT c : "clibos"
int CheckStackSpace();

// SPIN OBJECT strsub : "strsubs"
#include "strsubs.h"

// SPIN OBJECT
void ErrorExit(void);
void ProcessObject(int num, char **tokens);
void DumpSymbols(void);

char pins[] = {0, 0, 0, 0};
int currentwd = 0;
int locaddr1 = 0;

/* INLINE SPIN
PUB start(argc, argv)
  c.enter
  strsub.start
  main(argc, argv)
  waitcnt(clkfreq/10+cnt)
  c.exit(0)
*/

typedef char CHAR;

#define TYPE_GLOBAL   1
#define TYPE_LOCAL    2
#define TYPE_CONSTANT 3
#define TYPE_DAT      4
#define TYPE_VAR      5
#define TYPE_OBJECT   6
#define TYPE_PUB      7

#define TAB   9
#define OBJECT_OFFSET 16

#define OP_NONE                      0
#define OP_UNSIGNED_OBJ_OFFSET       1
#define OP_UNSIGNED_VAR_OFFSET       2
#define OP_UNSIGNED_LOC_OFFSET       3
#define OP_DATA                      4
#define OP_SIGNED_JMP_OFFSET         5
#define OP_MEMORY_OPCODE_WRITE       6
#define OP_OBJ_CALL_PAIR             7
#define OP_SIGNED_OFFSET             8
#define OP_PACKED_LITERAL            9
#define OP_BYTE_LITERAL             10
#define OP_WORD_LITERAL             11
#define OP_NEAR_LONG_LITERAL        12
#define OP_LONG_LITERAL             13
#define OP_MEMORY_OPCODE            14
#define OP_MEMORY_OPCODE_READ       15
#define OP_COMPACT_VAR_OFFSET       16
#define OP_COMPACT_LOC_OFFSET       17

#define MAX_TOKENS    40
#define MAX_LINE_SIZE 100
#define MAX_SYMBOLS   140
#define SYM_BUF_SIZE  MAX_SYMBOLS * 9
#define MAX_PUBS      40

int FindSymbol(CHAR *str);

int checksum = 0;
int pc = 0;
int pc0 = 0;
FILE *infile;
FILE *outfile;
int listflag = 0;
int verbose = 1;
int numsym = 0;
int linenum = 0;
CHAR *bigptr;
char zero_char = 0;
static int numobj = 0;
static int objnum = 0;

typedef struct symbolS {
    CHAR *name;
    int value;
    int type;
} symbolT;

CHAR ops[3000];
char bigbuf[SYM_BUF_SIZE];
symbolT symbol[MAX_SYMBOLS];
char buffer[MAX_LINE_SIZE];
char buffer1[MAX_LINE_SIZE];
char buffer2[MAX_LINE_SIZE+MAX_TOKENS];
unsigned short header[MAX_PUBS*2+10];

// Memory access pseudo-ops
char *pseudo_ops[] = { "ldb", "stb", "exb", "lab",
    "ldbx", "stbx", "exbx", "labx", "ldw",
    "stw", "exw", "law", "ldwx", "stwx",
    "exwx", "lawx", "ldl", "stl", "exl",
    "lal", "ldlx", "stlx", "exlx", "lalx", 0};

void ReadOpsFile(void)
{
    CHAR *ptr = ops;
    int i = 0;

#ifdef CSPIN_FLAG
    infile = fopen("/bin/opsfile.txt", "r");
#else
    infile = fopen("opsfile.txt", "r");
#endif

    if (!infile)
    {
#ifdef CSPIN_FLAG
	printf("Could not open /bin/opsfile.txt\n");
#else
	printf("Could not open opsfile.txt\n");
#endif
	exit(1);
    }

    while (fgets(ptr, 100, infile))
    {
	RemoveCRLF(ptr);
	//printf("ops %d: %s\n", i, ptr);
	ptr += strlen(ptr) + 1;
	i++;
    }
    *ptr = 0;
    fclose(infile);
}

int GetSymbolValue(CHAR *str, int pass);

int IsConstant(CHAR *str)
{
    int j = *str;
    if (j == '$' || j == DOUBLE_QUOTE || j == '%')
        return 1;
    if (j == '-'|| (j >= '0' && j <= '9'))
        return 1;
    if ((j = FindSymbol(str)) >= 0)
    {
        j = symbol[j].type;
        if (j == TYPE_GLOBAL || j == TYPE_CONSTANT)
            return 1;
    }
    return 0;
}

int GetDecVal(char *str)
{
    int signflag, val;
    int retval = 0;

    if (*str == '-')
    {
        signflag = 1;
        str++;
    }
    else
        signflag = 0;
    while (val = *str++)
    {
        if (val == '_')
            continue;
        if (val >= '0' && val <= '9')
            retval = (retval * 10) + val - '0';
        else
        {
            printf("Invalid number\n");
            ErrorExit();
        }
    }
    if (signflag)
        retval = -retval;
    return retval;
}

int GetHexVal(char *str)
{
    int val;
    int retval = 0;

    while (val = *str++)
    {
        if (val == '_')
            continue;
        if (val >= '0' && val <= '9')
            val -= '0';
        else if (val >= 'a' && val <= 'f')
            val -= 'a' - 10;
        else if (val >= 'A' && val <= 'F')
            val -= 'A' - 10;
        else
        {
            printf("Invalid number\n");
            ErrorExit();
        }
        retval = (retval << 4) + val;
    }

    return retval;
}

int GetBinVal(char *str)
{
    int val;
    int retval = 0;

    while (val = *str++)
    {
        if (val == '_')
            continue;
        if ((val & 0xFE) == '0')
          retval = (retval << 1) | (val & 1);
        else
        {
            printf("Invalid number\n");
            ErrorExit();
        }
    }

    return retval;
}

int GetVal(CHAR *str, int pass)
{
    int val;
    if (*str == '$')
        val = GetHexVal(str+1);
    else if(*str == DOUBLE_QUOTE)
        val = str[1];
    else if(*str == '%')
        val = GetBinVal(str+1);
    else if (*str == '-' || (*str >= '0' && *str <= '9'))
        val = GetDecVal(str);
    else
    {
	val = GetSymbolValue(str, pass);
    }
    return val;
}

void WriteByte(int value, int pass)
{
    if (pass != 2) return;
    if (listflag) printf(" %02x", value&255);
    fwrite(&value, 1, 1, outfile);
    checksum = (checksum + value) & 255;
}

int GetOpsFormat(char *str)
{
    int j;

    j = str[1];

    if (j == ' ') j = 0;
    else j -= '@';

    if (str[0] != ' ') j |= (str[0] - '0') << 5;

    return j;
}

CHAR *FindOp(CHAR *str, int *ops_format, int *ops_code, int pass)
{
    int i = 0;
    CHAR *ops_str = ops;

    while (*ops_str)
    {
        if (strcmp(ops_str + 6, str) == 0)
	    break;
	i++;
	ops_str += strlen(ops_str) + 1;
    }

    if(!(*ops_str))
    {
	if (pass == 2) printf("ERROR: Invalid opcode   %s\n", buffer1);
    }
    else
    {
        *ops_format = GetOpsFormat(ops_str);
        sscanf(ops_str+2, "%x", ops_code);
    }

    return ops_str;
}

CHAR *FindOpcode(int ops_code, int *ops_format)
{
    int temp;
    CHAR *ops_str = ops;

    if (ops_code >= 0x40 && ops_code < 0x80) ops_code &= 0xe3;

    while (*ops_str)
    {
        sscanf(ops_str+2, "%x", &temp);
	if (ops_code == temp) break;
	ops_str += strlen(ops_str) + 1;
    }

    if(*ops_str)
        *ops_format = GetOpsFormat(ops_str);

    return ops_str;
}

int ProcessEffect(CHAR **tokens, int tokenum, int pass, int num)
{
    CHAR *ops_str;
    int ops_format;
    int ops_code;

    ops_str = FindOp(tokens[tokenum], &ops_format, &ops_code, pass);
    if(!(*ops_str)) return 0;
    if (ops_code >= 0xe0) ops_code = (ops_code & 0x1f) + 0x40;
    else if((ops_format >> 5) != 3)
    {
        if (pass == 2) printf("ERROR: Invalid opcode   %s\n", buffer1);
        return 0;
    }

    if (strcmp(tokens[tokenum+1], "load") == 0) ops_code |= 0x80;

    WriteByte(ops_code, pass);
    return num + 1;
}

void ProcessDat(CHAR **tokens, int numtokens, int pass)
{
    char *ptr;
    int i, j, val, num, mode;
    int listflag1 = listflag && pass == 2;

    // Look for byte, word or long keyword
    if (StrCompNoCase(tokens[0], "byte")) num = 1;
    else if (StrCompNoCase(tokens[0], "word")) num = 2;
    else if (StrCompNoCase(tokens[0], "long")) num = 4;
    else return;

    if (listflag1) printf("%04x ", pc + OBJECT_OFFSET);

    // Add zero bytes for alignment
    if (num == 2)
    {
	if (pc&1)
	{
	    WriteByte(0, pass);
	    pc++;
            if (listflag1) printf("\n%04x ", pc + OBJECT_OFFSET);
	}
    }
    else if (num == 4)
    {
	if (pc&3)
	{
	    while (pc&3)
	    {
	        WriteByte(0, pass);
	        pc++;
	    }
            if (listflag1) printf("\n%04x ", pc + OBJECT_OFFSET);
	}
    }

    // Generate the output bytes
    j = 0;
    mode = 0;
    while (1)
    {
	if (mode == 0)
	{
	    j++;
	    if (j >= 10) break;
	    if (tokens[j][0] == 0) break;
	    if (tokens[j][0] == ',') continue;
	    if (tokens[j][0] == DOUBLE_QUOTE)
	    {
		ptr = &tokens[j][1];
		mode = 1;
	    }
	    else
	        val = GetVal(tokens[j], pass);
	}
	if (mode == 1)
	{
	    if (*ptr == 0 || *ptr == DOUBLE_QUOTE)
	    {
	        mode = 0;
		continue;
	    }
	    else
	    {
		val = *ptr++;
	    }
	}
	if (num == 1)
	{
            WriteByte(val&255, pass);
	}
	else if (num == 2)
	{
            WriteByte(val&255, pass);
            WriteByte((val>>8)&255, pass);
	}
	else
	{
            WriteByte(val&255, pass);
            WriteByte((val>>8)&255, pass);
            WriteByte((val>>16)&255, pass);
            WriteByte((val>>24)&255, pass);
	}
	pc += num;
    }

    if (listflag1)
    {
        for (i = num;i < 6; i++)
	{
	    printf("   ");
	}
        printf(" %s\n", buffer1);
    }
}

void Process(CHAR **tokens, int numtokens, int numpub, int pass)
{
    int i=0;
    int j;
    int val;
    int val2;
    int num = 0;
    int listflag1 = listflag && pass == 2;
    CHAR *ptr;
    int ops_code;
    int ops_format;
    CHAR *ops_str;
    int tokenum = 0;
    int index;

    //ops_str = ops;

    //printf("Process -%s- -%s-\n", tokens[0], tokens[1]);
    // Check for a pseudo-op
    index = FindListNoCase(tokens[0], pseudo_ops);

    if (index >= 0)
    {
	if (IsConstant(tokens[1]))
	{
            val = GetVal(tokens[1], pass);
	    if (val == -1)
	        ops_code = 0x34;
	    else if (val == 0)
	        ops_code = 0x35;
	    else if (val == 1)
	        ops_code = 0x36;
	    else if (val < 0 || val >= 0x01000000)
	        ops_code = 0x3b;
	    else if (val >= 0x00010000)
	        ops_code = 0x3a;
	    else if (val >= 0x00000100)
	        ops_code = 0x39;
	    else
	        ops_code = 0x38;
	}
	else if (tokens[1][0] == 0)
	{
	    ops_code = 0x80 | ((index & 0x1c) << 2) | (index & 3);
	}
	else
	{
            j = FindSymbol(tokens[1]);
            if (j < 0)
            {
	        printf("Symbol %s is undefined\n", tokens[1]);
                ErrorExit();
            }
            val = symbol[j].value;
            if (symbol[j].type == TYPE_LOCAL)
	    {
		if (val < 0x20 && (index & 0x1c) == 0x10)
		{
		    ops_code = 0x60 | val | (index & 3);
		    //printf("..llc $%2.2x $%2.2x\n", val, ops_code);
		}
		else
	            ops_code = 0x8c | ((index & 0x1c) << 2) | (index & 3);
	    }
	    else if (symbol[j].type == TYPE_DAT)
	    {
	        ops_code = 0x84 | ((index & 0x1c) << 2) | (index & 3);
	    }
	    else if (symbol[j].type == TYPE_VAR)
	    {
		if (val < 0x20 && (index & 0x1c) == 0x10)
		    ops_code = 0x40 | val | (index & 3);
		else
	            ops_code = 0x88 | ((index & 0x1c) << 2) | (index & 3);
	    }
	    else
	    {
		printf("Invalid type\n");
                ErrorExit();
	    }
	}
        ops_str = FindOpcode(ops_code, &ops_format);
    }
    else
    {
        ops_str = FindOp(tokens[0], &ops_format, &ops_code, pass);
    }

    if(!(*ops_str)) return;

    if ((ops_format & 0x1f) == OP_DATA)
    {
        ProcessDat(tokens, numtokens, pass);
        return;
    }

    if (listflag1) printf("%04x ", pc + OBJECT_OFFSET);

    switch (ops_format & 0x1f)
    {
	case OP_NONE:
	    WriteByte(ops_code, pass);
	    num = 1;
	    break;
#if 0
	case OP_DATA:
            val = GetVal(tokens[1], pass);
	    num = ops_code;
	    if (num == 1)
	    {
	        WriteByte(val&255, pass);
	    }
	    else if (num == 2)
	    {
		if (pc&1)
		{
		    WriteByte(0, pass);
		    pc++;
                    if (listflag1) printf("\n%04x ", pc + OBJECT_OFFSET);
		}
	        WriteByte(val&255, pass);
	        WriteByte((val>>8)&255, pass);
	    }
	    else
	    {
		if (pc&3)
		{
		    while (pc&3)
		    {
		        WriteByte(0, pass);
		        pc++;
		    }
                    if (listflag1) printf("\n%04x ", pc + OBJECT_OFFSET);
		}
	        WriteByte(val&255, pass);
	        WriteByte((val>>8)&255, pass);
	        WriteByte((val>>16)&255, pass);
	        WriteByte((val>>24)&255, pass);
	    }
	    break;
#endif
	case OP_BYTE_LITERAL:
            val = GetVal(tokens[1], pass);
	    WriteByte(ops_code, pass);
	    WriteByte(val&255, pass);
	    num = 2;
	    break;
	case OP_WORD_LITERAL:
            val = GetVal(tokens[1], pass);
	    WriteByte(ops_code, pass);
	    WriteByte((val>>8)&255, pass);
	    WriteByte(val&255, pass);
	    num = 3;
	    break;
	case OP_NEAR_LONG_LITERAL:
            val = GetVal(tokens[1], pass);
	    WriteByte(ops_code, pass);
	    WriteByte((val>>16)&255, pass);
	    WriteByte((val>>8)&255, pass);
	    WriteByte(val&255, pass);
	    num = 4;
	    break;
	case OP_LONG_LITERAL:
            val = GetVal(tokens[1], pass);
	    WriteByte(ops_code, pass);
	    WriteByte((val>>24)&255, pass);
	    WriteByte((val>>16)&255, pass);
	    WriteByte((val>>8)&255, pass);
	    WriteByte(val&255, pass);
	    num = 5;
	    break;
#ifdef CSPIN_FLAG
	case OP_COMPACT_VAR_OFFSET, OP_COMPACT_LOC_OFFSET:
#else
	case OP_COMPACT_VAR_OFFSET:
	case OP_COMPACT_LOC_OFFSET:
#endif
            val = GetVal(tokens[1], pass);
            WriteByte(ops_code | (val & 0x1c), pass);
	    num = 1;
	    tokenum++;
	    break;
	case OP_SIGNED_JMP_OFFSET:
            val = GetVal(tokens[1], pass);
#if 0
	    val -= pc + 2;
	    if (val >= -64 && val <= 63)
	    {
	        WriteByte(ops_code, pass);
	        WriteByte(val&127, pass);
	        num = 2;
	    }
	    else
	    {
		val--;
		printf("ERROR ");
	        WriteByte(ops_code, pass);
	        WriteByte(((val>>8)&255)|128, pass);
	        WriteByte(val&255, pass);
	        num = 3;
	    }
#else
	    val -= pc + 3;
            WriteByte(ops_code, pass);
            WriteByte(((val>>8)&255)|128, pass);
            WriteByte(val&255, pass);
            num = 3;
#endif
	    break;
#ifdef CSPIN_FLAG
	case OP_SIGNED_OFFSET, OP_UNSIGNED_OBJ_OFFSET, OP_UNSIGNED_VAR_OFFSET, OP_UNSIGNED_LOC_OFFSET:
#else
	case OP_SIGNED_OFFSET:
	case OP_UNSIGNED_OBJ_OFFSET:
	case OP_UNSIGNED_VAR_OFFSET:
	case OP_UNSIGNED_LOC_OFFSET:
#endif
            val = GetVal(tokens[1], pass);
	    //val -= 16;
#if 0
	    if (val >= -64 && val <= 63)
	    {
	        WriteByte(ops_code, pass);
	        WriteByte(val&127, pass);
	        num = 2;
	    }
	    else
	    {
		val--;
		printf("ERROR ");
	        WriteByte(ops_code, pass);
	        WriteByte(((val>>8)&255)|128, pass);
	        WriteByte(val&255, pass);
	        num = 3;
	    }
#else
            WriteByte(ops_code, pass);
            WriteByte(((val>>8)&255)|128, pass);
            WriteByte(val&255, pass);
            num = 3;
#endif
            tokenum++;
	    break;
        case OP_OBJ_CALL_PAIR:
            if (numtokens == 3)
            {
                val = GetVal(tokens[1], pass);
                val2 = GetVal(tokens[2], pass);
            }
            else
            {
                strcpy(buffer, tokens[1]);
                ptr = FindChar(buffer, '.');
                *ptr = 0;
                val = GetVal(buffer, pass) + numpub;
                val2 = GetVal(tokens[1], pass);
            }
	    WriteByte(ops_code, pass);
	    WriteByte(val&255, pass);
	    WriteByte(val2&255, pass);
	    num = 3;
	    break;
        case OP_MEMORY_OPCODE_READ:
            val = GetVal(tokens[1], pass);
	    WriteByte(ops_code, pass);
	    WriteByte(0x80|(val&31), pass);
	    num = 2;
	    break;
        case OP_MEMORY_OPCODE_WRITE:
            val = GetVal(tokens[1], pass);
	    WriteByte(ops_code, pass);
	    WriteByte(0xA0|(val&31), pass);
	    num = 2;
	    break;
        //case OP_PACKED_LITERAL:
        //case OP_MEMORY_OPCODE:
	default:
            if (pass == 2) printf("NOT IMPLEMENTED");
	    break;
    }

    if ((ops_format >> 5) == 1)
    {
	num = ProcessEffect(tokens, tokenum+1, pass, num);
	if (num == 0) return;
    }

    if (listflag1)
    {
        for (i = num;i < 6; i++)
	{
	    printf("   ");
	}
        printf(" %s\n", buffer1);
    }
    pc += num;
}

int FindSymbol(CHAR *str)
{
    int i;

    for (i = 0; i < numsym; i++)
    {
	if (symbol[i].type == 0) continue;
	if (StrCompNoCase(symbol[i].name, str)) break;
    }
    if (i < numsym) return i;
    return -1;
}

int GetSymbolValue(CHAR *str, int pass)
{
    int i;

    for (i = 0; i < numsym; i++)
    {
	if (symbol[i].type == 0) continue;
	if (strcmp(symbol[i].name, str) == 0) break;
    }
    if (i < numsym)
        return symbol[i].value;

    if (pass == 2)
    {
        printf("%s is undefined\n", str);
        ErrorExit();
    }

    return 0;
}

int AddSymbol(CHAR *str, int value, int type)
{
    int i;

    //printf("AddSymbol: %s, %x, %d\n", str, value, type);
  
    // Re-use symbol if previously used
    i = 0;
    while (i < numsym)
    {
        if (symbol[i].type == 0)
        {
            if (!strcmp(str, symbol[i].name))
                break;
        }
        i++;
    }
  
    // Otherwise, allocate new symbol entry
    if (i == numsym)
    {
        if (numsym >= MAX_SYMBOLS)
        {
            printf("Too many symbols\n");
            ErrorExit();
        }
        if (bigptr + strlen(str) + 1 > bigbuf + SYM_BUF_SIZE)
        {
            printf("Symbol buffer is full\n");
            ErrorExit();
        }
        strcpy(bigptr, str);
        //c.printf1(string("Adding to position %d\n"), i)
        symbol[i].name = bigptr;
        bigptr += strlen(bigptr) + 1;
        numsym++;
    }

    symbol[i].value = value;
    symbol[i].type = type;

    return i;
}

void RemoveLocalSymbols(void)
{
    int i;

    for (i = 0; i < numsym; i++)
    {
	if (symbol[i].type == TYPE_LOCAL)
	    symbol[i].type = 0;
    }
}

void PrintSymbolTable(void)
{
    int i;
    printf("\nSymbol Table\n");
    for (i = 0; i < numsym; i++)
    {
        printf("%s $%x (%d) %d\n",
	    symbol[i].name, symbol[i].value, symbol[i].value, symbol[i].type);
    }
}

void ProcessConstants(CHAR **tokens, int numtokens, int type)
{
    AddSymbol(tokens[0], GetVal(tokens[2], 0), type);
}

void usage(void)
{
    printf("usage: spasm [-l] [-d] file\n");
    exit(1);
}

void PrintHeader(int pubnum, int numobj)
{
    int i;

    checksum = 0;

    pc = (pc + 3 + 1) & 0xfffc;

    i = GetSymbolValue("_clkfreq", 0);
    if (verbose >= 2)
        printf("PrintHeader: _clkfreq = %d\n", i);
    header[0] = i & 0xffff;
    header[1] = (i >> 16) & 0xffff;
    header[2] = GetSymbolValue("_clkmode", 0); // sum-clk
    header[3] = OBJECT_OFFSET; // object offset
    header[4] = pc + OBJECT_OFFSET;
    header[5] = pc + 8 + OBJECT_OFFSET;
    header[6] = header[10] + OBJECT_OFFSET; // pc0;
    header[7] = pc + 8 + OBJECT_OFFSET + locaddr1;
    header[8] = pc;
    header[9] = (numobj << 8) | pubnum;

    if (verbose >= 2)
        printf("PrintHeader: numobj = %d\n", numobj);

#if 0
    for (i = 0; i < 8 + (pubnum*2); i++)
    {
        printf("header[%2d] = %04x\n", i, header[i]);
	checksum += header[i];
	checksum += header[i] >> 8;
    }
    fwrite(header, 1, (pubnum*4) + 16, outfile);
#else
    for (i = 0; i < 8 + ((pubnum+numobj)*2); i++)
    {
        if (listflag && i == 8) printf("\n");
        if (listflag) printf("%04x ", i << 1);
        WriteByte(header[i], 2);
        WriteByte(header[i] >> 8, 2);
        if (listflag) printf("\n");
    }
#endif
#if 0
    for (i = 0; i < numobj * 2; i++)
    {
        if (listflag && i == 8) printf("\n");
        if (listflag) printf("%04x ", i << 1);
        WriteByte(0, 2);
        WriteByte(0, 2);
        if (listflag) printf("\n");
    }
#endif
}

void PrintTail(void)
{
    int i;
    int num = 3;

    if (listflag) printf("\n%04x ", pc + OBJECT_OFFSET);
    while ((pc&3) != 3)
    {
	WriteByte(0, 2);
	pc++;
	num--;
    }
    WriteByte((0x14 - checksum)&255, 2);
    if (listflag)
    {
        for (i = 0; i < num; i++)
	{
	    printf("   ");
	}
        printf("               Trailing bytes\n");
    }
}

void AddStandardSymbols(void)
{
    AddSymbol("codid", 0x1e9, TYPE_CONSTANT);
    AddSymbol("par", 0x1f0, TYPE_CONSTANT);
    AddSymbol("cnt", 0x1f1, TYPE_CONSTANT);
    AddSymbol("ina", 0x1f2, TYPE_CONSTANT);
    AddSymbol("inb", 0x1f3, TYPE_CONSTANT);
    AddSymbol("outa", 0x1f4, TYPE_CONSTANT);
    AddSymbol("outb", 0x1f5, TYPE_CONSTANT);
    AddSymbol("dira", 0x1f6, TYPE_CONSTANT);
    AddSymbol("dirb", 0x1f7, TYPE_CONSTANT);
    AddSymbol("ctra", 0x1f8, TYPE_CONSTANT);
    AddSymbol("ctrb", 0x1f9, TYPE_CONSTANT);
    AddSymbol("frqa", 0x1fa, TYPE_CONSTANT);
    AddSymbol("frqb", 0x1fb, TYPE_CONSTANT);
    AddSymbol("phsa", 0x1fc, TYPE_CONSTANT);
    AddSymbol("phsb", 0x1fd, TYPE_CONSTANT);
    AddSymbol("vcfg", 0x1fe, TYPE_CONSTANT);
    AddSymbol("vscl", 0x1ff, TYPE_CONSTANT);
    AddSymbol("xtal1_pll16x", 0x6f, TYPE_CONSTANT);
    AddSymbol("xinput", 0x22, TYPE_CONSTANT);
    AddSymbol("negx", 0x80000000, TYPE_CONSTANT);
    AddSymbol("posx", 0x7fffffff, TYPE_CONSTANT);
}

void ProcessPub(CHAR **tokens, int numtokens, int pubnum, int pass)
{
    int i;
    int locmode, locspace, locaddr;

    //printf("ProcessPub: %s\n", tokens[1]);

    RemoveLocalSymbols();
    AddSymbol("result", 0, TYPE_LOCAL);
    if (pass == 1)
    {
	AddSymbol(tokens[1], pubnum, TYPE_GLOBAL);
        header[8+(pubnum<<1)] = pc;
    }
    locmode = locspace = 0;
    locaddr = 4;
    for (i = 2; i < numtokens; i++)
    {
        if (tokens[i][0] == 0) break;
        if (tokens[i][0] == SINGLE_QUOTE) break;
        if (tokens[i][0] == '|') locmode = 1;
        else if(tokens[i][0] != '(' && tokens[i][0] != ')' && tokens[i][0] != ',')
        {
	    AddSymbol(tokens[i], locaddr, TYPE_LOCAL);
	    locaddr += 4;
	    if (locmode) locspace += 4;
        }
    }
    if (pass == 1)
    {
        header[9+(pubnum<<1)] = locspace;
        if (pubnum == 1)
            locaddr1 = locaddr;
    }
}

void ErrorExit(void)
{
    printf("%d: %s\n", linenum, buffer);
    DumpSymbols();
    fclose(infile);
    exit(1);
}

void AddExternals(char **tokens, int numtokens)
{
    int i;
    char *ptr;

    for (i = 0; i < numtokens; i++)
    {
        if (tokens[i][0] == SINGLE_QUOTE)
            continue;
        if (tokens[i][0] == DOUBLE_QUOTE)
            continue;
        ptr = FindChar(tokens[i], '.');
        if (*ptr)
        {
            if (FindSymbol(tokens[i]) < 0)
	        AddSymbol(tokens[i], -1, TYPE_PUB);
            continue;
        }
        ptr = FindChar(tokens[i], '#');
        if (*ptr)
        {
            if (FindSymbol(tokens[i]) < 0)
	        AddSymbol(tokens[i], -1, TYPE_CONSTANT);
            continue;
        }
    }
}

int main(int argc, CHAR **argv)
{
    CHAR *tokens[MAX_TOKENS];
    int mode = 1;
    int pubnum = 1;
    int numpub = 0;
    int i;
    CHAR *fname = 0;
    int numtokens;
    CHAR **tokenptr;
    char *ptr;

    bigptr = bigbuf;

    ReadOpsFile();

    for (i = 1; i < argc; i++)
    {
        if (!strcmp(argv[i], "-l"))
            listflag = 1;
        else if (!strcmp(argv[i], "-d"))
            verbose = 2;
        else if (fname == 0)
            fname = argv[i];
        else
            usage();
    }
    if (fname == 0) usage();

    strcpy(buffer, fname);
    ptr = FindChar(buffer, '.');
    if (!(*ptr))
        strcpy(ptr, ".spa");
    infile = fopen(buffer, "r");
    if (!infile)
    {
        printf("Could not open input file %s\n", buffer);
        usage();
    }
    strcpy(ptr, ".bin");
    outfile = fopen(buffer, "wb");
    if (!outfile)
    {
        printf("Could not open output file %s\n", buffer);
        usage();
    }

    AddStandardSymbols();

    // Pass 1 - Compute starting address by counting pubs and objects
    if (verbose)
        printf("Pass 1: Counting methods and objects\n");

    pc0 = 4;
    linenum = 0;
    while (fgets(buffer, 100, infile))
    {
	RemoveCRLF(buffer);
	numtokens = GetTokens(buffer, buffer2, tokens);
        AddExternals(tokens, numtokens);
        linenum++;
        if (numtokens < 1) continue;
	if (StrCompNoCase(tokens[0], "pri"))
	    mode = TYPE_PUB;
	else if (StrCompNoCase(tokens[0], "obj"))
	    mode = TYPE_OBJECT;
	else if (StrCompNoCase(tokens[0], "dat"))
	    mode = TYPE_DAT;
	else if (StrCompNoCase(tokens[0], "var"))
	    mode = TYPE_VAR;
	else if (StrCompNoCase(tokens[0], "con"))
	    mode = TYPE_CONSTANT;
	else if (StrCompNoCase(tokens[0], "pub"))
        {
	    mode = TYPE_PUB;
            numpub++;
        }
        else if (mode == TYPE_OBJECT)
        {
            ptr = FindChar(buffer, ':');
            if (*ptr)
            {
                numobj++;
                if (verbose >= 2) printf("Count object - %s\n", buffer);
            }
        }
    }
    pc0 = ((numpub + numobj) << 2) + 4;

    if (verbose >= 2)
    {
        printf("pc0 = %d\n", pc0);
        printf("%d methods found\n", numpub);
        printf("%d objects found\n", numobj);
    }
    fseek(infile, 0, SEEK_SET);

    // Pass 2 - Compile Symbols
    if (verbose)
        printf("Pass 2: Compiling Symbols\n");

    pc = pc0;
    mode = 1;
    linenum = 0;
    while (fgets(buffer, 100, infile))
    {
        linenum++;
	RemoveCRLF(buffer);
	//tokenize(buffer, buffer2, tokens);
	numtokens = GetTokens(buffer, buffer2, tokens);
	for (i = numtokens; i < MAX_TOKENS; i++) tokens[i] = &zero_char;
        tokenptr = tokens;
	if (tokens[0][0] == 0) continue;
        if (tokens[0][0] == SINGLE_QUOTE) continue;
	if (buffer[0] != ' ' && buffer[0] != TAB)
	{
	    if (StrCompNoCase(tokens[0], "con"))
	        mode = 0;
	    else if (StrCompNoCase(tokens[0], "dat"))
	    {
	        mode = 2;
		continue;
	    }
	    else if (StrCompNoCase(tokens[0], "var"))
	    {
	        mode = 3;
		continue;
	    }
	    else if (StrCompNoCase(tokens[0], "pub"))
	    {
		ProcessPub(tokens, numtokens, pubnum, 1);
                pubnum++;
	        mode = 1;
	    }
	    else if (StrCompNoCase(tokens[0], "obj"))
	    {
	        mode = 4;
                continue;
	    }
	    else if (mode == 2)
            {
                if (numtokens >= 2 && StrCompNoCase(tokens[1], "long"))
                    AddSymbol(tokens[0], (((pc + 3) >> 2) << 2), TYPE_DAT);
                else if (numtokens >= 2 && StrCompNoCase(tokens[1], "word"))
                    AddSymbol(tokens[0], (((pc + 1) >> 1) << 1), TYPE_DAT);
                else
	            AddSymbol(tokens[0], pc, TYPE_DAT);
            }
	    else if (mode == 3)
            {
                if (StrCompNoCase(tokens[0], "long"))
                    AddSymbol(tokens[1], (((pc + 3) >> 2) << 2), TYPE_VAR);
                else if (StrCompNoCase(tokens[0], "word"))
                    AddSymbol(tokens[1], (((pc + 1) >> 1) << 1), TYPE_VAR);
                else
	            AddSymbol(tokens[1], pc, TYPE_VAR);
            }
	    else
	        AddSymbol(tokens[0], pc, TYPE_GLOBAL);
            if (numtokens == 1) continue;
            numtokens--;
#ifdef CSPIN_FLAG
            tokenptr += 4;
#else
            tokenptr++;
#endif
	}
	if (mode == 1)
	    Process(tokenptr, numtokens, numpub, 1);
	else if (mode == 2)
	    ProcessDat(tokenptr, numtokens, 1);
	else if (mode == 0)
	    ProcessConstants(tokenptr, numtokens, 1);
        else if (mode == 4)
	    ProcessObject(numtokens, tokenptr);
    }

    // Pass 3
    if (verbose)
    {
        printf("%d symbols compiled\n", numsym);
        printf("Pass 3: Generating bytecodes\n");
    }

    PrintHeader(pubnum, numobj);
    //PrintSymbolTable();

    fseek(infile, 0, SEEK_SET);
    pc = pc0;
    mode = 1;
    linenum = 0;
    while (fgets(buffer, 100, infile))
    {
        linenum++;
	RemoveCRLF(buffer);
	strcpy(buffer1, buffer);
	//tokenize(buffer, buffer2, tokens);
	numtokens = GetTokens(buffer, buffer2, tokens);
        tokenptr = tokens;
	for (i = numtokens; i < MAX_TOKENS; i++) tokens[i] = &zero_char;
        if (tokens[0][0] == SINGLE_QUOTE)
	{
            if (listflag) printf("                       %s\n", buffer1);
	    continue;
	}
	if (buffer[0] != ' ' && buffer[0] != TAB)
	{
	    if (StrCompNoCase(tokens[0], "con"))
	    {
	        mode = 0;
                if (listflag) printf("                        %s\n", buffer1);
		continue;
	    }
	    if (StrCompNoCase(tokens[0], "pub"))
	    {
                if (listflag) printf("                        %s\n", buffer1);
		ProcessPub(tokens, numtokens, pubnum, 2);
	        mode = 1;
		continue;
	    }
	    if (StrCompNoCase(tokens[0], "dat"))
	    {
	        mode = 2;
                if (listflag) printf("                        %s\n", buffer1);
		continue;
	    }
	    if (StrCompNoCase(tokens[0], "obj"))
	    {
	        mode = 4;
                if (listflag) printf("                        %s\n", buffer1);
		continue;
	    }
	    if (tokens[1][0] == 0)
	    {
                if (listflag) printf("                        %s\n", buffer1);
	        continue;
	    }
            numtokens--;
#ifdef CSPIN_FLAG
            tokenptr += 4;
#else
            tokenptr++;
#endif
	}
	if (mode == 1)
	    Process(tokenptr, numtokens, numpub, 2);
	else if (mode == 2)
	    ProcessDat(tokenptr, numtokens, 2);
	else
            if (listflag) printf("                        %s\n", buffer1);
    }
    PrintTail();
    fclose(infile);
    fclose(outfile);
    if (verbose)
        printf("%d bytes generated\n", pc + OBJECT_OFFSET + 1);
    if (verbose >= 2)
        DumpSymbols();
}

// *********************************************************************
// This routine is used during the first passed to extract
// the constants and public methods from an external object
// and add them to the symbol table.
// *********************************************************************
void ProcessObject(int num, char **tokens)
{
    int len, val;
    int entrynum;
    int entrynum1;
    FILE *objfile;
    char objname[20];
    int mode = TYPE_CONSTANT;
    int numpub = 1;

    if (num < 1 || tokens[0][0] == SINGLE_QUOTE) return;
    if (num != 3 || strcmp(tokens[1], ":"))
    {
	printf("Error in object line\n");
        ErrorExit();
    }
    if (verbose >= 2)
        printf("ProcessObject: AddSymbol %s\n", tokens[0]);
    entrynum = AddSymbol(tokens[0], ++objnum, TYPE_GLOBAL);
    strcpy(objname, tokens[2] + 1);
    len = strlen(objname);
    if (len >= 1) objname[len-1] = 0;
    strcat(objname, ".spa");
    if (verbose >= 2)
        printf("ProcessObject %s %s %s\n", tokens[0], tokens[1], tokens[2]);
    objfile = fopen(objname, "r");
    if (!objfile)
    {
	printf("Could not open object file %s\n", tokens[2]);
        ErrorExit();
    }

    while (fgets(buffer, 100, objfile))
    {
	RemoveCRLF(buffer);
	num = GetTokens(buffer, buffer2, tokens);
	if (StrCompNoCase(tokens[0], "pub"))
	{
	    mode = TYPE_PUB;
	    if (num >= 2)
	    {
		strcpy(objname, symbol[entrynum].name);
		strcat(objname, ".");
		strcat(objname, tokens[1]);
                entrynum1 = FindSymbol(objname);
                if (entrynum1 >= 0)
                {
                    if (verbose >= 2)
		        printf("ProcessObject: Set value of %s to %d\n", objname, numpub);
                    symbol[entrynum1].value = numpub;
                }
                else
                {
		    //printf("ProcessObject: Skip adding %s\n", objname);
                }
                numpub++;
	    }
	}
	else if (StrCompNoCase(tokens[0], "pri"))
	    mode = TYPE_PUB;
	else if (StrCompNoCase(tokens[0], "obj"))
	    mode = TYPE_OBJECT;
	else if (StrCompNoCase(tokens[0], "dat"))
	    mode = TYPE_DAT;
	else if (StrCompNoCase(tokens[0], "var"))
	    mode = TYPE_VAR;
	else if (StrCompNoCase(tokens[0], "con"))
	    mode = TYPE_CONSTANT;
	else if (mode == TYPE_CONSTANT)
	{
	    if (num >= 2 && strcmp(tokens[1], "=") == 0)
	    {
		strcpy(objname, symbol[entrynum].name);
		strcat(objname, "#");
		strcat(objname, tokens[0]);
                entrynum1 = FindSymbol(objname);
                if (entrynum1 >= 0)
                {
                    val = GetVal(tokens[2], 0);
                    if (verbose >= 2)
		        printf("ProcessObject: Set value of %s to %d\n", objname, val);
                    symbol[entrynum1].value = val;
                }
                else
                {
		    //printf("ProcessObject: Skip adding %s\n", objname);
                }
	    }
	}
    }
    fclose(objfile);
}

void DumpSymbols(void)
{
    int i;

    printf("Symbol Table\n");
    for (i = 0; i < numsym; i++)
        printf("Symbol %d: %s %d %d\n", i, symbol[i].name, symbol[i].value, symbol[i].type);
}
/*
+-----------------------------------------------------------------------------+
|                       TERMS OF USE: MIT License                             |
+-----------------------------------------------------------------------------+
|Permission is hereby granted, free of charge, to any person obtaining a copy |
|of this software and associated documentation files (the "Software"), to deal|
|in the Software without restriction, including without limitation the rights |
|to use, copy, modify, merge, publish, distribute, sublicense, and/or sell    |
|copies of the Software, and to permit persons to whom the Software is        |
|furnished to do so, subject to the following conditions:                     |
|                                                                             |
|The above copyright notice and this permission notice shall be included in   |
|all copies or substantial portions of the Software.                          |
|                                                                             |
|THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR   |
|IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,     |
|FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE  |
|AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER       |
|LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,|
|OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE|
|SOFTWARE.                                                                    |
+-----------------------------------------------------------------------------+
*/
